from unittest import mock

from django.test import SimpleTestCase, override_settings, tag

from anymail.backends.base_requests import AnymailRequestsBackend, RequestsPayload
from anymail.message import AnymailMessage, AnymailRecipientStatus
from tests.utils import AnymailTestMixin

from .mock_requests_backend import RequestsBackendMockAPITestCase


class MinimalRequestsBackend(AnymailRequestsBackend):
    """(useful only for these tests)"""

    esp_name = "Example"
    api_url = "https://httpbin.org/post"  # helpful echoback endpoint for live testing

    def __init__(self, **kwargs):
        super().__init__(self.api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        _payload_init = getattr(message, "_payload_init", {})
        return MinimalRequestsPayload(message, defaults, self, **_payload_init)

    def parse_recipient_status(self, response, payload, message):
        return {"to@example.com": AnymailRecipientStatus("message-id", "sent")}


class MinimalRequestsPayload(RequestsPayload):
    def init_payload(self):
        pass

    def _noop(self, *args, **kwargs):
        pass

    set_from_email = _noop
    set_recipients = _noop
    set_subject = _noop
    set_reply_to = _noop
    set_extra_headers = _noop
    set_text_body = _noop
    set_html_body = _noop
    add_attachment = _noop


@override_settings(EMAIL_BACKEND="tests.test_base_backends.MinimalRequestsBackend")
class RequestsBackendBaseTestCase(RequestsBackendMockAPITestCase):
    """Test common functionality in AnymailRequestsBackend"""

    def setUp(self):
        super().setUp()
        self.message = AnymailMessage(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    def test_minimal_requests_backend(self):
        """Make sure the testing backend defined above actually works"""
        self.message.send()
        self.assert_esp_called("https://httpbin.org/post")

    def test_timeout_default(self):
        """All requests have a 30 second default timeout"""
        self.message.send()
        timeout = self.get_api_call_arg("timeout")
        self.assertEqual(timeout, 30)

    @override_settings(ANYMAIL_REQUESTS_TIMEOUT=5)
    def test_timeout_setting(self):
        """You can use the Anymail setting REQUESTS_TIMEOUT to override the default"""
        self.message.send()
        timeout = self.get_api_call_arg("timeout")
        self.assertEqual(timeout, 5)

    @mock.patch(f"{__name__}.MinimalRequestsBackend.create_session")
    def test_create_session_error_fail_silently(self, mock_create_session):
        # If create_session fails and fail_silently is True,
        # make sure _send doesn't raise a misleading error.
        mock_create_session.side_effect = ValueError("couldn't create session")
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)


@tag("live")
@override_settings(EMAIL_BACKEND="tests.test_base_backends.MinimalRequestsBackend")
class RequestsBackendLiveTestCase(AnymailTestMixin, SimpleTestCase):
    @override_settings(ANYMAIL_DEBUG_API_REQUESTS=True)
    def test_debug_logging(self):
        message = AnymailMessage(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )
        message._payload_init = dict(
            data="Request body",
            headers={
                "Content-Type": "text/plain",
                "Accept": "application/json",
            },
        )
        with self.assertPrints("===== Anymail API request") as outbuf:
            message.send()

        # Header order and response data vary too much to do a full comparison,
        # but make sure that the output contains some expected pieces of the request
        # and the response
        output = outbuf.getvalue()
        self.assertIn("\nPOST https://httpbin.org/post\n", output)
        self.assertIn("\nUser-Agent: django-anymail/", output)
        self.assertIn("\nAccept: application/json\n", output)
        self.assertIn("\nContent-Type: text/plain\n", output)  # request
        self.assertIn("\n\nRequest body\n", output)
        self.assertIn("\n----- Response\n", output)
        self.assertIn("\nHTTP 200 OK\n", output)
        self.assertIn("\nContent-Type: application/json\n", output)  # response

    def test_no_debug_logging(self):
        # Make sure it doesn't output anything when DEBUG_API_REQUESTS is not set
        message = AnymailMessage(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )
        message._payload_init = dict(
            data="Request body",
            headers={
                "Content-Type": "text/plain",
                "Accept": "application/json",
            },
        )
        with self.assertPrints("", match="equal"):
            message.send()



import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import tag

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.brevo import BrevoTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase


@tag("brevo")
class BrevoWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    def call_webhook(self):
        return self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps({}),
        )

    # Actual tests are in WebhookBasicAuthTestCase


@tag("brevo")
class BrevoDeliveryTestCase(WebhookTestCase):
    # Brevo's webhook payload data is documented at
    # https://developers.brevo.com/docs/transactional-webhooks.
    # The payloads below were obtained through live testing.

    def test_sent_event(self):
        raw_event = {
            "event": "request",
            "email": "recipient@example.com",
            "id": 9999999,  # this seems to be Brevo account id (not an event id)
            "message-id": "<201803062010.27287306012@smtp-relay.mailin.fr>",
            "subject": "Test subject",
            # From a message sent at 2018-03-06 11:10:23-08:00
            # (2018-03-06 19:10:23+00:00)...
            "date": "2018-03-06 11:10:23",  # tz from Brevo account's preferences
            "ts": 1520331023,  # 2018-03-06 10:10:23 -- what time zone is this?
            "ts_event": 1520331023,  # unclear if this ever differs from "ts"
            "ts_epoch": 1520363423000,  # 2018-03-06 19:10:23.000+00:00 -- UTC (msec)
            "X-Mailin-custom": '{"meta": "data"}',
            # "tag" is JSON-serialized tags array if `tags` param set on send,
            #   else single tag string if `X-Mailin-Tag` header set on send,
            #   else template name if sent using a template,
            #   else not present.
            # "tags" is tags list if `tags` param set on send, else not present.
            "tag": '["tag1","tag2"]',
            "tags": ["tag1", "tag2"],
            "template_id": 12,
            "sending_ip": "333.33.33.33",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "queued")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 6, 19, 10, 23, microsecond=0, tzinfo=timezone.utc),
        )
        self.assertEqual(
            event.message_id, "<201803062010.27287306012@smtp-relay.mailin.fr>"
        )
        # Brevo does not provide a unique event id:
        self.assertIsNone(event.event_id)
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.metadata, {"meta": "data"})
        self.assertEqual(event.tags, ["tag1", "tag2"])

    def test_delivered_event(self):
        raw_event = {
            # For brevity, this and following tests omit some webhook data
            # that was tested earlier, or that is not used by Anymail
            "event": "delivered",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.message_id, "<201803011158.9876543210@smtp-relay.mailin.fr>"
        )
        self.assertEqual(event.recipient, "recipient@example.com")
        # empty dict when no X-Mailin-custom header given:
        self.assertEqual(event.metadata, {})
        self.assertEqual(event.tags, [])  # empty list when no tags given

    def test_hard_bounce(self):
        raw_event = {
            "event": "hard_bounce",
            "email": "not-a-user@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            # the leading space in the reason is as received in actual testing:
            "reason": " RecipientError: 550 5.5.0"
            " Requested action not taken: mailbox unavailable.",
            "tag": "header-tag",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(
            event.mta_response,
            " RecipientError: 550 5.5.0"
            " Requested action not taken: mailbox unavailable.",
        )
        self.assertEqual(event.tags, ["header-tag"])

    def test_soft_bounce_event(self):
        raw_event = {
            "event": "soft_bounce",
            "email": "recipient@no-mx.example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            "reason": "undefined Unable to find MX of domain no-mx.example.com",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "bounced")
        # no human-readable description consistently available:
        self.assertIsNone(event.description)
        self.assertEqual(
            event.mta_response,
            "undefined Unable to find MX of domain no-mx.example.com",
        )

    def test_blocked(self):
        raw_event = {
            "event": "blocked",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            "reason": "blocked : due to blacklist user",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.reject_reason, "blocked")
        self.assertEqual(event.mta_response, "blocked : due to blacklist user")

    def test_spam(self):
        # "When a person who received your email reported that it is a spam."
        # (haven't observed "spam" event in actual testing; payload below is a guess)
        raw_event = {
            "event": "spam",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.reject_reason, "spam")

    def test_complaint(self):
        # Sadly, this is not well documented in the official Brevo API documentation.
        raw_event = {
            "event": "complaint",
            "email": "example@domain.com",
            "id": "xxxxx",
            "date": "2020-10-09 00:00:00",
            "ts": 1604933619,
            "message-id": "201798300811.5787683@relay.domain.com",
            "ts_event": 1604933654,
            "X-Mailin-custom": '{"meta": "data"}',
            "tags": ["transac_messages"],
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.reject_reason, "spam")

    def test_invalid_email(self):
        # "If a ISP again indicated us that the email is not valid or if we discovered
        # that the email is not valid." (unclear whether this error originates with the
        # receiving MTA or with Brevo pre-send) (haven't observed "invalid_email"
        # event in actual testing; payload below is a guess)
        raw_event = {
            "event": "invalid_email",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            "reason": "(guessing invalid_email includes a reason)",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "invalid")
        self.assertEqual(
            event.mta_response, "(guessing invalid_email includes a reason)"
        )

    def test_error_email(self):
        # Sadly, this is not well documented in the official Brevo API documentation.
        raw_event = {
            "event": "error",
            "email": "example@domain.com",
            "id": "xxxxx",
            "date": "2020-10-09 00:00:00",
            "ts": 1604933619,
            "message-id": "201798300811.5787683@relay.domain.com",
            "ts_event": 1604933654,
            "subject": "My first Transactional",
            "X-Mailin-custom": '{"meta": "data"}',
            "template_id": 22,
            "tags": ["transac_messages"],
            "ts_epoch": 1604933623,
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "failed")
        self.assertEqual(event.reject_reason, None)

    def test_deferred_event(self):
        # Note: the example below is an actual event capture (with 'example.com'
        # substituted for the real receiving domain). It's pretty clearly a bounce, not
        # a deferral. It looks like Brevo mis-categorizes this SMTP response code.
        raw_event = {
            "event": "deferred",
            "email": "notauser@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            "reason": "550 RecipientError: 550 5.1.1 <notauser@example.com>: Recipient"
            " address rejected: User unknown in virtual alias table",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        # no human-readable description consistently available:
        self.assertIsNone(event.description)
        self.assertEqual(
            event.mta_response,
            "550 RecipientError: 550 5.1.1 <notauser@example.com>:"
            " Recipient address rejected: User unknown in virtual alias table",
        )

    def test_opened_event(self):
        # Brevo delivers 'unique_opened' only on the first open, and 'opened'
        # only on the second or later tracking pixel views. (But they used to deliver
        # both on the first open.)
        raw_event = {
            "event": "opened",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertIsNone(event.user_agent)  # Brevo doesn't report user agent

    def test_unique_opened_event(self):
        # See note in test_opened_event above
        raw_event = {
            "event": "unique_opened",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")

    def test_proxy_open_event(self):
        # Equivalent to "Loaded via Proxy" in the Brevo UI.
        # This is sent when a tracking pixel is loaded via a 'privacy proxy server'.
        # This technique is used by Apple Mail, for example, to protect user's IP
        # addresses.
        raw_event = {
            "event": "proxy_open",
            "email": "example@domain.com",
            "id": 1,
            "date": "2020-10-09 00:00:00",
            "message-id": "201798300811.5787683@relay.domain.com",
            "subject": "My first Transactional",
            "tag": ["transactionalTag"],
            "sending_ip": "xxx.xxx.xxx.xxx",
            "s_epoch": 1534486682000,
            "template_id": 1,
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")

    def test_unique_proxy_open_event(self):
        # Sadly, undocumented in Brevo.
        # Equivalent to "First Open but loaded via Proxy".
        # This is sent when a tracking pixel is loaded via a 'privacy proxy server'.
        # This technique is used by Apple Mail, for example, to protect user's IP
        # addresses.
        raw_event = {
            "event": "unique_proxy_open",
            "email": "example@domain.com",
            "id": 1,
            "date": "2020-10-09 00:00:00",
            "message-id": "201798300811.5787683@relay.domain.com",
            "subject": "My first Transactional",
            "tag": ["transactionalTag"],
            "sending_ip": "xxx.xxx.xxx.xxx",
            "s_epoch": 1534486682000,
            "template_id": 1,
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")

    def test_clicked_event(self):
        raw_event = {
            "event": "click",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
            "link": "https://example.com/click/me",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.click_url, "https://example.com/click/me")
        self.assertIsNone(event.user_agent)  # Brevo doesn't report user agent

    def test_unsubscribe(self):
        # "When a person unsubscribes from the email received."
        # (haven't observed "unsubscribe" event in actual testing;
        # payload below is a guess)
        raw_event = {
            "event": "unsubscribe",
            "email": "recipient@example.com",
            "ts_epoch": 1520363423000,
            "message-id": "<201803011158.9876543210@smtp-relay.mailin.fr>",
        }
        response = self.client.post(
            "/anymail/brevo/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=BrevoTrackingWebhookView,
            event=ANY,
            esp_name="Brevo",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unsubscribed")

    def test_misconfigured_inbound(self):
        errmsg = (
            "You seem to have set Brevo's *inbound* webhook URL"
            " to Anymail's Brevo *tracking* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/brevo/tracking/",
                content_type="application/json",
                data={"items": []},
            )



from base64 import b64encode
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import (
    AnymailAPIError,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("postal")
@override_settings(
    EMAIL_BACKEND="anymail.backends.postal.EmailBackend",
    ANYMAIL={
        "POSTAL_API_KEY": "test_server_token",
        "POSTAL_API_URL": "https://postal.example.com",
    },
)
class PostalBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""{
        "status": "success",
        "time": 2.82,
        "flags": {},
        "data": {
            "message_id": "ad5084a6-cf01-448b-92da-2574ee64c0ba@rp.postal.example.com",
            "messages": {
                "to@example.com": { "id": 1503, "token": "ke47R2hZkSlA" }
            }
        }
    }"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("postal")
class PostalBackendStandardEmailTests(PostalBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/message")
        headers = self.get_api_call_headers()
        self.assertEqual(headers["X-Server-API-Key"], "test_server_token")
        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["plain_body"], "Here is the message.")
        self.assertEqual(data["from"], "from@sender.example.com")
        self.assertEqual(data["to"], ["to@example.com"])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(data["from"], "From Name <from@example.com>")
        self.assertEqual(
            data["to"], ["Recipient #1 <to1@example.com>", "to2@example.com"]
        )
        self.assertEqual(
            data["cc"], ["Carbon Copy <cc1@example.com>", "cc2@example.com"]
        )
        self.assertEqual(
            data["bcc"], ["Blind Copy <bcc1@example.com>", "bcc2@example.com"]
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                # should override backend msgid:
                "Message-ID": "mycustommsgid@sales.example.com",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject")
        self.assertEqual(data["plain_body"], "Body goes here")
        self.assertEqual(data["from"], "from@example.com")
        self.assertEqual(data["to"], ["to1@example.com", "Also To <to2@example.com>"])
        self.assertEqual(
            data["bcc"], ["bcc1@example.com", "Also BCC <bcc2@example.com>"]
        )
        self.assertEqual(data["cc"], ["cc1@example.com", "Also CC <cc2@example.com>"])
        self.assertEqual(data["reply_to"], "another@example.com")
        self.assertCountEqual(
            data["headers"],
            {"Message-ID": "mycustommsgid@sales.example.com", "X-MyHeader": "my value"},
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["plain_body"], text_content)
        self.assertEqual(data["html_body"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertNotIn("plain_body", data)
        self.assertEqual(data["html_body"], html_content)

    def test_extra_headers(self):
        self.message.extra_headers = {"X-Custom": "string", "X-Num": 123}
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(data["headers"], {"X-Custom": "string", "X-Num": 123})

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    @override_settings(ANYMAIL_IGNORE_UNSUPPORTED_FEATURES=True)
    def test_reply_to(self):
        # Postal only allows single reply-to. Test handling for multiple reply
        # addresses when ignoring errors:
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["reply_to"], "reply@example.com")  # keeps first email

    def test_multiple_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
        )
        with self.assertRaises(AnymailUnsupportedFeature):
            email.send()

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["name"], "test.txt")
        self.assertEqual(attachments[0]["content_type"], "text/plain")
        self.assertEqual(
            decode_att(attachments[0]["data"]).decode("ascii"), text_content
        )

        # content_type inferred from filename:
        self.assertEqual(attachments[1]["content_type"], "image/png")
        self.assertEqual(attachments[1]["name"], "test.png")
        self.assertEqual(decode_att(attachments[1]["data"]), png_content)

        self.assertEqual(attachments[2]["content_type"], "application/pdf")
        self.assertEqual(attachments[2]["name"], "")  # none
        self.assertEqual(decode_att(attachments[2]["data"]), pdf_content)

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachments"],
            [
                {
                    "name": "Une pièce jointe.html",
                    "content_type": "text/html",
                    "data": b64encode("<p>\u2019</p>".encode("utf-8")).decode("ascii"),
                }
            ],
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        with self.assertRaisesMessage(AnymailUnsupportedFeature, "inline attachments"):
            self.message.send()

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        image_data_b64 = b64encode(image_data).decode("ascii")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachments"],
            [
                {
                    "name": image_filename,  # the named one
                    "content_type": "image/png",
                    "data": image_data_b64,
                },
                {
                    "name": "",  # the unnamed one
                    "content_type": "image/png",
                    "data": image_data_b64,
                },
            ],
        )

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple html parts"):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty fields"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("Cc", data)
        self.assertNotIn("Bcc", data)
        self.assertNotIn("ReplyTo", data)

        # Test empty `to`--but send requires at least one recipient somewhere (like cc)
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("To", data)

    def test_api_failure(self):
        failure_response = b"""{
            "status": "error",
            "time": 0.0,
            "flags": {},
            "data": {
                "code": "ValidationError"
            }
        }"""
        self.set_mock_response(status_code=200, raw=failure_response)
        with self.assertRaisesMessage(AnymailAPIError, "Postal API response 200"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=200, raw=failure_response)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response:
        error_response = b"""{
            "status": "error",
            "time": 0.0,
            "flags": {},
            "data": {
                "code": "ValidationError"
            }
        }"""
        self.set_mock_response(status_code=200, raw=error_response)
        with self.assertRaisesMessage(AnymailAPIError, "ValidationError"):
            self.message.send()

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Ack! Bad proxy!")
        with self.assertRaisesMessage(AnymailAPIError, "Ack! Bad proxy!"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("postal")
class PostalBackendAnymailFeatureTests(PostalBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "anything@bounces.example.com"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["sender"], "anything@bounces.example.com")

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "metadata"):
            self.message.send()

    def test_send_at(self):
        self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "send_at"):
            self.message.send()

    def test_tags(self):
        self.message.tags = ["receipt"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tag"], "receipt")

        self.message.tags = ["receipt", "repeat-user"]
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple tags"):
            self.message.send()

    def test_track_opens(self):
        self.message.track_opens = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_opens"):
            self.message.send()

    def test_track_clicks(self):
        self.message.track_clicks = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_clicks"):
            self.message.send()

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("tag", data)

    def test_esp_extra(self):
        self.message.esp_extra = {
            "future_postal_option": "some-value",
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["future_postal_option"], "some-value")

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = b"""{
          "status": "success",
          "time": 1.08,
          "flags": {},
          "data": {
            "message_id": "9dfcc4df-09a6-4f1d-b535-0eb0a9f104a4@postal.example.com",
            "messages": {
              "to1@example.com": { "id": 1531, "token": "xLcafDRCVUFe" }
            }
          }
        }"""
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["Recipient <to1@example.com>"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, 1531)
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, 1531
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_without_to_attaches_anymail_status(self):
        """The anymail_status should be attached even if there are no `to` recipients"""
        response_content = b"""{
          "status": "success",
          "time": 1.08,
          "flags": {},
          "data": {
            "message_id": "9dfcc4df-09a6-4f1d-b535-0eb0a9f104a4@postal.example.com",
            "messages": {
              "cc@example.com": { "id": 1531, "token": "xLcafDRCVUFe" }
            }
          }
        }"""
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            cc=["cc@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, 1531)
        self.assertEqual(
            msg.anymail_status.recipients["cc@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["cc@example.com"].message_id, 1531
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.tags = [Decimal("19.99")]  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Postal", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("postal")
class PostalBackendRecipientsRefusedTests(PostalBackendMockAPITestCase):
    # Postal doesn't check email bounce or complaint lists at time of send --
    # it always just queues the message. You'll need to listen for the "rejected"
    # and "failed" events to detect refused recipients.
    pass


@tag("postal")
class PostalBackendSessionSharingTestCase(
    SessionSharingTestCases, PostalBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("postal")
@override_settings(EMAIL_BACKEND="anymail.backends.postal.EmailBackend")
class PostalBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bPOSTAL_API_KEY\b")
        self.assertRegex(errmsg, r"\bANYMAIL_POSTAL_API_KEY\b")



import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import tag
from django.utils.timezone import get_fixed_timezone

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.postmark import PostmarkTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase


@tag("postmark")
class PostmarkWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    def call_webhook(self):
        return self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps({}),
        )

    # Actual tests are in WebhookBasicAuthTestCase


@tag("postmark")
class PostmarkDeliveryTestCase(WebhookTestCase):
    def test_bounce_event(self):
        raw_event = {
            "RecordType": "Bounce",
            "ID": 901542550,
            "Type": "HardBounce",
            "TypeCode": 1,
            "ServerID": 23,
            "Name": "Hard bounce",
            "MessageID": "2706ee8a-737c-4285-b032-ccd317af53ed",
            "Description": "The server was unable to deliver your message"
            " (ex: unknown user, mailbox not found).",
            "Details": "smtp;550 5.1.1 The email account that you tried to reach"
            " does not exist.",
            "Email": "bounce@example.com",
            "From": "sender@example.com",
            "BouncedAt": "2016-04-27T16:28:50.3963933-04:00",
            "DumpAvailable": True,
            "Inactive": True,
            "CanActivate": True,
            "Subject": "Postmark event test",
            "Content": "...",
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(
                2016,
                4,
                27,
                16,
                28,
                50,
                microsecond=396393,
                tzinfo=get_fixed_timezone(-4 * 60),
            ),
        )
        self.assertEqual(event.message_id, "2706ee8a-737c-4285-b032-ccd317af53ed")
        self.assertEqual(event.event_id, "901542550")
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(
            event.description,
            "The server was unable to deliver your message"
            " (ex: unknown user, mailbox not found).",
        )
        self.assertEqual(
            event.mta_response,
            "smtp;550 5.1.1 The email account that you tried to reach does not exist.",
        )

    def test_delivered_event(self):
        raw_event = {
            "RecordType": "Delivery",
            "ServerId": 23,
            "MessageID": "883953f4-6105-42a2-a16a-77a8eac79483",
            "Recipient": "recipient@example.com",
            "Tag": "welcome-email",
            "Metadata": {
                "cohort": "2014-08-A",
                "userid": "12345",  # Postmark metadata is always converted to string
            },
            "DeliveredAt": "2014-08-01T13:28:10.2735393-04:00",
            "Details": "Test delivery webhook details",
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(
                2014,
                8,
                1,
                13,
                28,
                10,
                microsecond=273539,
                tzinfo=get_fixed_timezone(-4 * 60),
            ),
        )
        self.assertEqual(event.message_id, "883953f4-6105-42a2-a16a-77a8eac79483")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["welcome-email"])
        self.assertEqual(event.metadata, {"cohort": "2014-08-A", "userid": "12345"})

    def test_open_event(self):
        raw_event = {
            "RecordType": "Open",
            "FirstOpen": True,
            "Client": {"Name": "Gmail", "Company": "Google", "Family": "Gmail"},
            "OS": {"Name": "unknown", "Company": "unknown", "Family": "unknown"},
            "Platform": "Unknown",
            "UserAgent": "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0",
            "ReadSeconds": 0,
            "Geo": {},
            "MessageID": "f4830d10-9c35-4f0c-bca3-3d9b459821f8",
            "ReceivedAt": "2016-04-27T16:21:41.2493688-04:00",
            "Recipient": "recipient@example.com",
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(
                2016,
                4,
                27,
                16,
                21,
                41,
                microsecond=249368,
                tzinfo=get_fixed_timezone(-4 * 60),
            ),
        )
        self.assertEqual(event.message_id, "f4830d10-9c35-4f0c-bca3-3d9b459821f8")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent, "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0"
        )
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_click_event(self):
        raw_event = {
            "RecordType": "Click",
            "ClickLocation": "HTML",
            "Client": {
                "Name": "Chrome 35.0.1916.153",
                "Company": "Google",
                "Family": "Chrome",
            },
            "OS": {
                "Name": "OS X 10.7 Lion",
                "Company": "Apple Computer, Inc.",
                "Family": "OS X 10",
            },
            "Platform": "Desktop",
            "UserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) etc.",
            "OriginalLink": "https://example.com/click/me",
            "Geo": {
                "CountryISOCode": "RS",
                "Country": "Serbia",
                "RegionISOCode": "VO",
                "Region": "Autonomna Pokrajina Vojvodina",
                "City": "Novi Sad",
                "Zip": "21000",
                "Coords": "45.2517,19.8369",
                "IP": "8.8.8.8",
            },
            "MessageID": "f4830d10-9c35-4f0c-bca3-3d9b459821f8",
            "ReceivedAt": "2017-10-25T15:21:11.9065619Z",
            "Tag": "welcome-email",
            "Recipient": "recipient@example.com",
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(2017, 10, 25, 15, 21, 11, microsecond=906561, tzinfo=timezone.utc),
        )
        self.assertEqual(event.message_id, "f4830d10-9c35-4f0c-bca3-3d9b459821f8")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent, "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) etc."
        )
        self.assertEqual(event.click_url, "https://example.com/click/me")
        self.assertEqual(event.tags, ["welcome-email"])
        self.assertEqual(event.metadata, {})

    def test_spam_event(self):
        raw_event = {
            "RecordType": "SpamComplaint",
            "ID": 901542550,
            "Type": "SpamComplaint",
            "TypeCode": 512,
            "Name": "Spam complaint",
            "Tag": "Test",
            "MessageID": "2706ee8a-737c-4285-b032-ccd317af53ed",
            "ServerID": 1234,
            "Description": "",
            "Details": "Test spam complaint details",
            "Email": "spam@example.com",
            "From": "sender@example.com",
            "BouncedAt": "2016-04-27T16:28:50.3963933-04:00",
            "DumpAvailable": True,
            "Inactive": True,
            "CanActivate": False,
            "Subject": "Postmark event test",
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp,
            datetime(
                2016,
                4,
                27,
                16,
                28,
                50,
                microsecond=396393,
                tzinfo=get_fixed_timezone(-4 * 60),
            ),
        )
        self.assertEqual(event.message_id, "2706ee8a-737c-4285-b032-ccd317af53ed")
        self.assertEqual(event.event_id, "901542550")
        self.assertEqual(event.recipient, "spam@example.com")
        self.assertEqual(event.reject_reason, "spam")
        self.assertEqual(event.description, "")
        self.assertEqual(event.mta_response, "Test spam complaint details")

    def test_misconfigured_inbound(self):
        errmsg = (
            "You seem to have set Postmark's *inbound* webhook"
            " to Anymail's Postmark *tracking* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/postmark/tracking/",
                content_type="application/json",
                data=json.dumps({"FromFull": {"Email": "from@example.org"}}),
            )

        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/postmark/tracking/",
                content_type="application/json",
                data=json.dumps({"RecordType": "Inbound"}),
            )

    def test_unsubscribe(self):
        raw_event = {
            "RecordType": "SubscriptionChange",
            "MessageID": "a4909a96-73d7-4c49-b148-a54522d3f7ac",
            "ServerID": 23,
            "MessageStream": "outbound",
            "ChangedAt": "2022-06-05T17:17:32Z",
            "Recipient": "john@example.com",
            "Origin": "Recipient",
            "SuppressSending": True,
            "SuppressionReason": "ManualSuppression",
            "Tag": "welcome-email",
            "Metadata": {"example": "value", "example_2": "value"},
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime(2022, 6, 5, 17, 17, 32, tzinfo=timezone.utc)
        )
        self.assertEqual(event.message_id, "a4909a96-73d7-4c49-b148-a54522d3f7ac")
        self.assertEqual(
            event.recipient,
            "john@example.com",
        )
        self.assertEqual(event.reject_reason, "unsubscribed")

    def test_resubscribe(self):
        raw_event = {
            "RecordType": "SubscriptionChange",
            "MessageID": "a4909a96-73d7-4c49-b148-a54522d3f7ac",
            "ServerID": 23,
            "MessageStream": "outbound",
            "ChangedAt": "2022-06-05T17:17:32Z",
            "Recipient": "john@example.com",
            "Origin": "Recipient",
            "SuppressSending": False,
            "SuppressionReason": None,
            "Tag": "welcome-email",
            "Metadata": {"example": "value", "example_2": "value"},
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "subscribed")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime(2022, 6, 5, 17, 17, 32, tzinfo=timezone.utc)
        )
        self.assertEqual(event.message_id, "a4909a96-73d7-4c49-b148-a54522d3f7ac")
        self.assertEqual(
            event.recipient,
            "john@example.com",
        )
        self.assertEqual(event.reject_reason, None)

    def test_subscription_change_bounce(self):
        raw_event = {
            "RecordType": "SubscriptionChange",
            "MessageID": "b4cb783d-78ed-43f2-983b-63f55c712dc8",
            "ServerID": 23,
            "MessageStream": "outbound",
            "ChangedAt": "2022-06-05T17:17:32Z",
            "Recipient": "john@example.com",
            "Origin": "Recipient",
            "SuppressSending": True,
            "SuppressionReason": "HardBounce",
            "Tag": "my-tag",
            "Metadata": {"example": "value", "example_2": "value"},
        }
        response = self.client.post(
            "/anymail/postmark/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostmarkTrackingWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime(2022, 6, 5, 17, 17, 32, tzinfo=timezone.utc)
        )
        self.assertEqual(event.message_id, "b4cb783d-78ed-43f2-983b-63f55c712dc8")
        self.assertEqual(
            event.recipient,
            "john@example.com",
        )
        self.assertEqual(event.reject_reason, "bounced")



import datetime

from django.core import mail
from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    decode_att,
    sample_image_content,
    sample_image_path,
)

# Minimal required ANYMAIL settings for Scaleway, used in multiple tests
SCALEWAY_BASE_SETTINGS = {
    "SCALEWAY_SECRET_KEY": "test_secret_key",
    "SCALEWAY_PROJECT_ID": "test_project_id",
}


@tag("scaleway")
@override_settings(
    EMAIL_BACKEND="anymail.backends.scaleway.EmailBackend",
    ANYMAIL=SCALEWAY_BASE_SETTINGS,
)
class ScalewayBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""{
        "emails": [
            {
                "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                "message_id": "00000000-1111-2222-3333-444444444444",
                "project_id": "test-project-id",
                "mail_from": "from@example.com",
                "mail_rcpt": "to@example.com",
                "rcpt_type": "to",
                "status": "sending",
                "status_details": "ready to send",
                "created_at": "2025-08-05T01:12:20.016801Z",
                "updated_at": "2025-08-05T01:12:20.016801Z"
            }
        ]
    }"""
    DEFAULT_CONTENT_TYPE = "application/json"

    def setUp(self):
        super().setUp()
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("scaleway")
class ScalewayBackendStandardEmailTests(ScalewayBackendMockAPITestCase):
    def test_send_mail(self):
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called(
            "https://api.scaleway.com/transactional-email/v1alpha1/regions/fr-par/emails"
        )
        headers = self.get_api_call_headers()
        self.assertEqual(headers["X-Auth-Token"], "test_secret_key")
        data = self.get_api_call_json()
        self.assertEqual(data["project_id"], "test_project_id")
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["text"], "Here is the message.")
        self.assertEqual(data["from"], {"email": "from@example.com"})
        self.assertEqual(data["to"], [{"email": "to@example.com"}])

    def test_name_addr(self):
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ['"Recipient, #1" <to1@example.com>', "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["from"], {"name": "From Name", "email": "from@example.com"}
        )
        self.assertEqual(
            data["to"],
            [
                {"email": "to1@example.com", "name": "Recipient, #1"},
                {"email": "to2@example.com"},
            ],
        )
        self.assertEqual(
            data["cc"],
            [
                {"email": "cc1@example.com", "name": "Carbon Copy"},
                {"email": "cc2@example.com"},
            ],
        )
        self.assertEqual(
            data["bcc"],
            [
                {"email": "bcc1@example.com", "name": "Blind Copy"},
                {"email": "bcc2@example.com"},
            ],
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["text"], text_content)
        self.assertEqual(data["html"], html_content)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertNotIn("text", data)
        self.assertEqual(data["html"], html_content)

    def test_extra_headers(self):
        self.message.extra_headers = {"X-Custom": "string", "X-Num": 123}
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["additional_headers"],
            [{"key": "X-Custom", "value": "string"}, {"key": "X-Num", "value": "123"}],
        )

    def test_reply_to(self):
        self.message.reply_to = ["reply@example.com"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["additional_headers"],
            [{"key": "Reply-To", "value": "reply@example.com"}],
        )

    def test_reply_to_and_extra_headers(self):
        self.message.reply_to = ["reply@example.com"]
        self.message.extra_headers = {"X-Custom": "string"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["additional_headers"],
            [
                {"key": "X-Custom", "value": "string"},
                {"key": "Reply-To", "value": "reply@example.com"},
            ],
        )

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["attachments"]), 1)
        self.assertEqual(data["attachments"][0]["name"], "test.txt")
        self.assertEqual(data["attachments"][0]["type"], "text/plain")
        self.assertEqual(
            decode_att(data["attachments"][0]["content"]).decode(), text_content
        )

    def test_inline_images(self):
        # Scaleway's API doesn't have a way to specify content-id
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = f'<p>This has an <img src="cid:{cid}" alt="inline" /> image.</p>'
        self.message.attach_alternative(html_content, "text/html")

        with self.assertRaisesMessage(AnymailUnsupportedFeature, "inline attachments"):
            self.message.send()

    @override_settings(ANYMAIL_IGNORE_UNSUPPORTED_FEATURES=True)
    def test_inline_images_ignore_unsupported(self):
        # Sends as ordinary attachment when ignoring unsupported features
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_content = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = f'<p>This has an <img src="cid:{cid}" alt="inline" /> image.</p>'
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["attachments"]), 1)
        self.assertEqual(data["attachments"][0]["name"], image_filename)
        self.assertEqual(data["attachments"][0]["type"], "image/png")
        self.assertEqual(
            decode_att(data["attachments"][0]["content"]),
            image_content,
        )

    def test_api_failure(self):
        self.set_mock_response(
            status_code=400, json_data={"message": "Helpful diagnostics from Scaleway"}
        )
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful diagnostics from Scaleway"
        ):
            self.message.send()


@tag("scaleway")
class ScalewayBackendAnymailFeatureTests(ScalewayBackendMockAPITestCase):
    def test_envelope_sender(self):
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["additional_headers"]), 1)
        self.assertEqual(data["additional_headers"][0]["key"], "X-Metadata")
        self.assertJSONEqual(
            data["additional_headers"][0]["value"],
            {"user_id": "12345", "items": 6},
        )

    def test_send_at(self):
        self.message.send_at = datetime.datetime(
            2022, 5, 6, 7, 8, 9, tzinfo=datetime.timezone.utc
        )
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "send_at"):
            self.message.send()

    def test_tags(self):
        self.message.tags = ["receipt", "reorder test 12"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["additional_headers"]), 1)
        self.assertEqual(data["additional_headers"][0]["key"], "X-Tags")
        self.assertJSONEqual(
            data["additional_headers"][0]["value"],
            ["receipt", "reorder test 12"],
        )

    def test_headers_interaction(self):
        # Test four features that use custom headers don't clobber each other
        self.message.reply_to = ["reply@example.com"]
        self.message.extra_headers = {"X-Custom": "custom value"}
        self.message.metadata = {"user_id": "12345"}
        self.message.tags = ["receipt"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(
            data["additional_headers"],
            [
                {"key": "Reply-To", "value": "reply@example.com"},
                {"key": "X-Custom", "value": "custom value"},
                {"key": "X-Tags", "value": '["receipt"]'},
                {"key": "X-Metadata", "value": '{"user_id": "12345"}'},
            ],
        )

    def test_merge_data(self):
        self.message.merge_data = {"to@example.com": {"customer_id": "3"}}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "merge_data"):
            self.message.send()

    def test_merge_global_data(self):
        self.message.merge_global_data = {"customer_id": "3"}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "merge_global_data"):
            self.message.send()

    def test_merge_metadata(self):
        self.message.merge_metadata = {"to@example.com": {"tier": "premium"}}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "merge_metadata"):
            self.message.send()

    def test_template_id(self):
        self.message.template_id = "template-12345"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "template_id"):
            self.message.send()

    def test_track_opens(self):
        self.message.track_opens = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_opens"):
            self.message.send()

    def test_track_clicks(self):
        self.message.track_clicks = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_clicks"):
            self.message.send()

    def test_esp_extra(self):
        self.message.esp_extra = {"send_before": "2022-05-06T07:08:09Z"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["send_before"], "2022-05-06T07:08:09Z")

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["Recipient <to1@example.com>"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        # Anymail's message_id is Scaleway's technical "id", which is unique per
        # recipient. (Not Scaleway's "message_id", which is shared across recipients.)
        self.assertEqual(
            msg.anymail_status.message_id,
            "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
        )
        self.assertEqual(
            msg.anymail_status.recipients["to@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to@example.com"].message_id,
            "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
        )
        self.assertEqual(
            msg.anymail_status.esp_response.content, self.DEFAULT_RAW_RESPONSE
        )

    def test_default_omits_unused_fields(self):
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("additional_headers", data)
        self.assertNotIn("attachments", data)
        self.assertNotIn("cc", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("html", data)


@tag("scaleway")
class ScalewayBackendRecipientsRefusedTests(ScalewayBackendMockAPITestCase):
    # Scaleway doesn't appear to check email bounce or complaint lists at time
    # of send -- it always just queues the message. You'll need to listen for
    # tracking webhook events to detect failed sends.
    pass


@tag("resend")
class ScalewayBackendSessionSharingTestCase(
    SessionSharingTestCases, ScalewayBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("scaleway")
@override_settings(EMAIL_BACKEND="anymail.backends.scaleway.EmailBackend")
class ScalewayBackendConfigurationTests(SimpleTestCase):
    @override_settings(ANYMAIL={"SCALEWAY_PROJECT_ID": "test_project_id"})
    def test_missing_secret_key(self):
        with self.assertRaisesRegex(
            AnymailConfigurationError, r"You must set.*SCALEWAY_SECRET_KEY"
        ):
            mail.get_connection()

    @override_settings(ANYMAIL={"SCALEWAY_SECRET_KEY": "test_secret_key"})
    def test_missing_project_id(self):
        with self.assertRaisesRegex(
            AnymailConfigurationError, r"You must set.*SCALEWAY_PROJECT_ID"
        ):
            mail.get_connection()

    @override_settings(
        ANYMAIL={"SCALEWAY_PROJECT_ID": "test_project_id"},
        SCALEWAY_SECRET_KEY="test_secret_key",
    )
    def test_bare_secret_key(self):
        # SCALEWAY_SECRET_KEY is allowed in settings file root
        connection = mail.get_connection()
        self.assertEqual(connection.secret_key, "test_secret_key")

    @override_settings(ANYMAIL=dict(SCALEWAY_BASE_SETTINGS, SCALEWAY_REGION="pl-waw"))
    def test_region_setting(self):
        backend = mail.get_connection()
        self.assertEqual(
            backend.api_url,
            "https://api.scaleway.com/transactional-email/v1alpha1/regions/pl-waw/",
        )

    @override_settings(
        ANYMAIL=dict(SCALEWAY_BASE_SETTINGS, SCALEWAY_REGION="nl-ams # from /.env")
    )
    def test_bad_region_setting(self):
        # Make sure api_url is properly quoted
        # (e.g., misread from an env file that doesn't allow comments)
        backend = mail.get_connection()
        self.assertEqual(
            backend.api_url,
            "https://api.scaleway.com/transactional-email/v1alpha1"
            "/regions/nl-ams%20%23%20from%20%2F.env/",
        )

    @override_settings(
        ANYMAIL=dict(
            SCALEWAY_BASE_SETTINGS,
            SCALEWAY_API_URL="https://scaleway.example.com/{region}/email",
        )
    )
    def test_api_url_setting(self):
        # The {region} placeholder is replaced with the region setting
        backend = mail.get_connection()
        self.assertEqual(backend.api_url, "https://scaleway.example.com/fr-par/email/")

    @override_settings(
        ANYMAIL=dict(
            SCALEWAY_BASE_SETTINGS,
            SCALEWAY_API_URL="https://scaleway.example.com/email",
        )
    )
    def test_api_url_setting_without_region(self):
        # The region placeholder is not required
        backend = mail.get_connection()
        self.assertEqual(backend.api_url, "https://scaleway.example.com/email/")



import warnings

from ..exceptions import AnymailDeprecationWarning
from .brevo import BrevoInboundWebhookView, BrevoTrackingWebhookView


class SendinBlueTrackingWebhookView(BrevoTrackingWebhookView):
    """
    Deprecated compatibility tracking webhook for old Brevo name "SendinBlue".
    """

    esp_name = "SendinBlue"

    def __init__(self, **kwargs):
        warnings.warn(
            "Anymail's SendinBlue webhook URLs are deprecated."
            " Update your Brevo transactional email webhook URL to change"
            " 'anymail/sendinblue' to 'anymail/brevo'.",
            AnymailDeprecationWarning,
        )
        super().__init__(**kwargs)


class SendinBlueInboundWebhookView(BrevoInboundWebhookView):
    """
    Deprecated compatibility inbound webhook for old Brevo name "SendinBlue".
    """

    esp_name = "SendinBlue"

    def __init__(self, **kwargs):
        warnings.warn(
            "Anymail's SendinBlue webhook URLs are deprecated."
            " Update your Brevo inbound webhook URL to change"
            " 'anymail/sendinblue' to 'anymail/brevo'.",
            AnymailDeprecationWarning,
        )
        super().__init__(**kwargs)



from django.core import mail

from ..exceptions import AnymailAPIError
from ..message import AnymailRecipientStatus
from .base import AnymailBaseBackend, BasePayload


class EmailBackend(AnymailBaseBackend):
    """
    Anymail backend that simulates sending messages, useful for testing.

    Sent messages are collected in django.core.mail.outbox
    (as with Django's locmem backend).

    In addition:
    * Anymail send params parsed from the message will be attached
      to the outbox message as a dict in the attr `anymail_test_params`
    * If the caller supplies an `anymail_test_response` attr on the message,
      that will be used instead of the default "sent" response. It can be either
      an AnymailRecipientStatus or an instance of AnymailAPIError (or a subclass)
      to raise an exception.
    """

    esp_name = "Test"

    def __init__(self, *args, **kwargs):
        # Allow replacing the payload, for testing.
        # (Real backends would generally not implement this option.)
        self._payload_class = kwargs.pop("payload_class", TestPayload)
        super().__init__(*args, **kwargs)
        if not hasattr(mail, "outbox"):
            mail.outbox = []  # see django.core.mail.backends.locmem

    def get_esp_message_id(self, message):
        # Get a unique ID for the message.  The message must have been added to
        # the outbox first.
        return mail.outbox.index(message)

    def build_message_payload(self, message, defaults):
        return self._payload_class(backend=self, message=message, defaults=defaults)

    def post_to_esp(self, payload, message):
        # Keep track of the sent messages and params (for test cases)
        message.anymail_test_params = payload.get_params()
        mail.outbox.append(message)
        try:
            # Tests can supply their own message.test_response:
            response = message.anymail_test_response
            if isinstance(response, AnymailAPIError):
                raise response
        except AttributeError:
            # Default is to return 'sent' for each recipient
            status = AnymailRecipientStatus(
                message_id=self.get_esp_message_id(message), status="sent"
            )
            response = {
                "recipient_status": {
                    email: status for email in payload.recipient_emails
                }
            }
        return response

    def parse_recipient_status(self, response, payload, message):
        try:
            return response["recipient_status"]
        except KeyError as err:
            raise AnymailAPIError("Unparsable test response") from err


class TestPayload(BasePayload):
    # For test purposes, just keep a dict of the params we've received.
    # (This approach is also useful for native API backends -- think of
    # payload.params as collecting kwargs for esp_native_api.send().)

    def init_payload(self):
        self.params = {}
        self.recipient_emails = []

    def get_params(self):
        # Test backend callers can check message.anymail_test_params['is_batch_send']
        # to verify whether Anymail thought the message should use batch send logic.
        self.params["is_batch_send"] = self.is_batch()
        return self.params

    def set_from_email(self, email):
        self.params["from"] = email

    def set_envelope_sender(self, email):
        self.params["envelope_sender"] = email.addr_spec

    def set_to(self, emails):
        self.params["to"] = emails
        self.recipient_emails += [email.addr_spec for email in emails]

    def set_cc(self, emails):
        self.params["cc"] = emails
        self.recipient_emails += [email.addr_spec for email in emails]

    def set_bcc(self, emails):
        self.params["bcc"] = emails
        self.recipient_emails += [email.addr_spec for email in emails]

    def set_subject(self, subject):
        self.params["subject"] = subject

    def set_reply_to(self, emails):
        self.params["reply_to"] = emails

    def set_extra_headers(self, headers):
        self.params["extra_headers"] = headers

    def set_text_body(self, body):
        self.params["text_body"] = body

    def set_html_body(self, body):
        self.params["html_body"] = body

    def add_alternative(self, content, mimetype):
        # For testing purposes, we allow all "text/*" alternatives,
        # but not any other mimetypes.
        if mimetype.startswith("text"):
            self.params.setdefault("alternatives", []).append((content, mimetype))
        else:
            self.unsupported_feature("alternative part with type '%s'" % mimetype)

    def add_attachment(self, attachment):
        self.params.setdefault("attachments", []).append(attachment)

    def set_metadata(self, metadata):
        self.params["metadata"] = metadata

    def set_send_at(self, send_at):
        self.params["send_at"] = send_at

    def set_tags(self, tags):
        self.params["tags"] = tags

    def set_track_clicks(self, track_clicks):
        self.params["track_clicks"] = track_clicks

    def set_track_opens(self, track_opens):
        self.params["track_opens"] = track_opens

    def set_template_id(self, template_id):
        self.params["template_id"] = template_id

    def set_merge_data(self, merge_data):
        self.params["merge_data"] = merge_data

    def set_merge_headers(self, merge_headers):
        self.params["merge_headers"] = merge_headers

    def set_merge_metadata(self, merge_metadata):
        self.params["merge_metadata"] = merge_metadata

    def set_merge_global_data(self, merge_global_data):
        self.params["merge_global_data"] = merge_global_data

    def set_esp_extra(self, extra):
        # Merge extra into params
        self.params.update(extra)



import json
from base64 import b64encode

from django.test import override_settings

from tests.utils import ClientWithCsrfChecks

HAS_CRYPTOGRAPHY = True
try:
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import ec
except ImportError:
    HAS_CRYPTOGRAPHY = False


def make_key() -> "ec.EllipticCurvePrivateKey":
    """Generate RSA public key with short key size, for testing only"""
    return ec.generate_private_key(
        curve=ec.SECP256R1(),
    )


def derive_public_webhook_key(private_key: "ec.EllipticCurvePrivateKey") -> str:
    """Derive public"""
    public_key = private_key.public_key()
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    public_bytes = b"\n".join(public_bytes.splitlines()[1:-1])
    return public_bytes.decode("utf-8")


def sign(
    private_key: "ec.EllipticCurvePrivateKey", timestamp: str, message: str
) -> bytes:
    """Sign message with private key"""
    return private_key.sign(
        (timestamp + message).encode("utf-8"), ec.ECDSA(hashes.SHA256())
    )


class _ClientWithSendGridSignature(ClientWithCsrfChecks):
    private_key = None

    def set_private_key(self, private_key):
        self.private_key = private_key

    def post(self, *args, **kwargs):
        # Timestamp will be a date string, but the exact value doesn't actually
        # matter for verification purposes
        timestamp = "timestamp"
        signature = b64encode(
            sign(
                self.private_key,
                timestamp=timestamp,
                message=json.dumps(kwargs["data"]),
            )
        )
        if kwargs.pop("automatically_set_timestamp_and_signature_headers", True):
            kwargs.setdefault("HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_TIMESTAMP", timestamp)
            kwargs.setdefault("HTTP_X_TWILIO_EMAIL_EVENT_WEBHOOK_SIGNATURE", signature)

        webhook_key = derive_public_webhook_key(self.private_key)
        with override_settings(
            ANYMAIL={"SENDGRID_TRACKING_WEBHOOK_VERIFICATION_KEY": webhook_key}
        ):
            return super().post(*args, **kwargs)


ClientWithSendGridSignature = _ClientWithSendGridSignature if HAS_CRYPTOGRAPHY else None



from datetime import date, datetime
from email import message_from_bytes
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from textwrap import dedent

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailError,
    AnymailInvalidAddress,
    AnymailRequestsAPIError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    sample_email_content,
    sample_image_content,
    sample_image_path,
)


@tag("mailgun")
@override_settings(
    EMAIL_BACKEND="anymail.backends.mailgun.EmailBackend",
    ANYMAIL={"MAILGUN_API_KEY": "test_api_key"},
)
class MailgunBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b"""{
        "id": "<20160306015544.116301.25145@example.com>",
        "message": "Queued. Thank you."
    }"""

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("mailgun")
class MailgunBackendStandardEmailTests(MailgunBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/example.com/messages")
        auth = self.get_api_call_auth()
        self.assertEqual(auth, ("api", "test_api_key"))
        data = self.get_api_call_data()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["text"], "Here is the message.")
        self.assertEqual(data["from"], ["from@example.com"])
        self.assertEqual(data["to"], ["to@example.com"])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_data()
        self.assertEqual(data["from"], ["From Name <from@example.com>"])
        self.assertEqual(
            data["to"], ["Recipient #1 <to1@example.com>", "to2@example.com"]
        )
        self.assertEqual(
            data["cc"], ["Carbon Copy <cc1@example.com>", "cc2@example.com"]
        )
        self.assertEqual(
            data["bcc"], ["Blind Copy <bcc1@example.com>", "bcc2@example.com"]
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "x-my-header": "my value",
                "Message-ID": "mycustommsgid@example.com",
            },
        )
        email.send()
        data = self.get_api_call_data()
        self.assertEqual(data["subject"], "Subject")
        self.assertEqual(data["text"], "Body goes here")
        self.assertEqual(data["from"], ["from@example.com"])
        self.assertEqual(data["to"], ["to1@example.com", "Also To <to2@example.com>"])
        self.assertEqual(
            data["bcc"], ["bcc1@example.com", "Also BCC <bcc2@example.com>"]
        )
        self.assertEqual(data["cc"], ["cc1@example.com", "Also CC <cc2@example.com>"])
        self.assertEqual(data["h:Reply-To"], "another@example.com")
        self.assertEqual(data["h:X-My-Header"], "my value")
        self.assertEqual(data["h:Message-Id"], "mycustommsgid@example.com")
        # multiple recipients, but not a batch send:
        self.assertNotIn("recipient-variables", data)

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_data()
        self.assertEqual(data["text"], text_content)
        self.assertEqual(data["html"], html_content)
        # Don't accidentally send the html part as an attachment:
        files = self.get_api_call_files(required=False)
        self.assertFalse(files)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_data()
        self.assertNotIn("text", data)
        self.assertEqual(data["html"], html_content)

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
            headers={"X-Other": "Keep"},
        )
        email.send()
        data = self.get_api_call_data()
        self.assertEqual(
            data["h:Reply-To"], "reply@example.com, Other <reply2@example.com>"
        )
        self.assertEqual(data["h:X-Other"], "Keep")  # don't lose other headers

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        # Mailgun requires filename:
        mimeattachment["Content-Disposition"] = 'attachment; filename="custom filename"'
        self.message.attach(mimeattachment)

        # And also with an message/rfc822 attachment
        forwarded_email_content = sample_email_content()
        forwarded_email = message_from_bytes(forwarded_email_content)
        rfcmessage = MIMEBase("message", "rfc822")
        # Mailgun requires filename:
        rfcmessage.add_header(
            "Content-Disposition", "attachment", filename="forwarded message"
        )
        rfcmessage.attach(forwarded_email)
        self.message.attach(rfcmessage)

        self.message.send()
        files = self.get_api_call_files()
        attachments = [value for (field, value) in files if field == "attachment"]
        self.assertEqual(len(attachments), 4)
        self.assertEqual(attachments[0], ("test.txt", text_content, "text/plain"))
        self.assertEqual(
            # (type inferred from filename)
            attachments[1],
            ("test.png", png_content, "image/png"),
        )
        self.assertEqual(
            attachments[2], ("custom filename", pdf_content, "application/pdf")
        )
        # Email messages can get a bit changed with respect to whitespace characters
        # in headers, without breaking the message, so we tolerate that:
        self.assertEqual(attachments[3][0], "forwarded message")
        self.assertEqualIgnoringHeaderFolding(
            attachments[3][1],
            b"Content-Type: message/rfc822\nMIME-Version: 1.0\n"
            + b'Content-Disposition: attachment; filename="forwarded message"\n'
            + b"\n"
            + forwarded_email_content,
        )
        self.assertEqual(attachments[3][2], "message/rfc822")

        # Make sure the image attachment is not treated as embedded:
        inlines = [value for (field, value) in files if field == "inline"]
        self.assertEqual(len(inlines), 0)

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()

        # Verify the RFC 7578 compliance workaround has kicked in:
        data = self.get_api_call_data()
        if isinstance(data, dict):
            # workaround not needed or used (but let's double check actual request)
            workaround = False
            prepared = self.get_api_prepared_request()
            data = prepared.body
        else:
            workaround = True
        data = data.decode("utf-8").replace("\r\n", "\n")
        self.assertNotIn("filename*=", data)  # No RFC 2231 encoding
        self.assertIn(
            'Content-Disposition: form-data; name="attachment";'
            ' filename="Une pièce jointe.html"',
            data,
        )

        if workaround:
            files = self.get_api_call_files(required=False)
            self.assertFalse(files)  # files should have been moved to formdata body

    def test_rfc_7578_compliance(self):
        # Check some corner cases in the workaround
        # that undoes RFC 2231 multipart/form-data encoding...
        self.message.subject = "Testing for filename*=utf-8''problems"
        self.message.body = (
            "The attached message should have an attachment named 'vedhæftet fil.txt'"
        )
        # A forwarded message with its own attachment:
        forwarded_message = dedent(
            """\
            MIME-Version: 1.0
            From: sender@example.com
            Subject: This is a test message
            Content-Type: multipart/mixed; boundary="boundary"

            --boundary
            Content-Type: text/plain

            This message has an attached file with a non-ASCII filename.
            --boundary
            Content-Type: text/plain; name*=utf-8''vedh%C3%A6ftet%20fil.txt
            Content-Disposition: attachment; filename*=utf-8''vedh%C3%A6ftet%20fil.txt

            This is an attachment.
            --boundary--
            """
        )
        self.message.attach(
            "besked med vedhæftede filer", forwarded_message, "message/rfc822"
        )
        self.message.send()

        data = self.get_api_call_data()
        if isinstance(data, dict):
            # workaround not needed or used (but let's double check actual request)
            prepared = self.get_api_prepared_request()
            data = prepared.body
        data = data.decode("utf-8").replace("\r\n", "\n")

        # Top-level attachment (in form-data)
        # should have RFC 7578 filename (raw Unicode):
        self.assertIn(
            'Content-Disposition: form-data; name="attachment";'
            ' filename="besked med vedhæftede filer"',
            data,
        )
        # Embedded message/rfc822 attachment should retain its RFC 2231 encoded
        # filename:
        self.assertIn(
            "Content-Type: text/plain; name*=utf-8''vedh%C3%A6ftet%20fil.txt", data
        )
        self.assertIn(
            "Content-Disposition: attachment;"
            " filename*=utf-8''vedh%C3%A6ftet%20fil.txt",
            data,
        )
        # References to RFC 2231 in message text should remain intact:
        self.assertIn("Testing for filename*=utf-8''problems", data)
        self.assertIn(
            "The attached message should have an attachment named 'vedhæftet fil.txt'",
            data,
        )

    def test_attachment_missing_filename(self):
        """Mailgun silently drops attachments without filenames, so warn the caller"""
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(b"PDF\xb4 pretend this is valid pdf data")
        mimeattachment["Content-Disposition"] = "attachment"
        self.message.attach(mimeattachment)

        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "attachments without filenames"
        ):
            self.message.send()

    def test_inline_missing_contnet_id(self):
        mimeattachment = MIMEImage(b"imagedata", "x-fakeimage")
        mimeattachment["Content-Disposition"] = 'inline; filename="fakeimage.txt"'
        self.message.attach(mimeattachment)
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "inline attachments without Content-ID"
        ):
            self.message.send()

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["html"], html_content)

        files = self.get_api_call_files()
        inlines = [value for (field, value) in files if field == "inline"]
        self.assertEqual(len(inlines), 1)
        # filename is cid; type is guessed:
        self.assertEqual(inlines[0], (cid, image_data, "image/png"))
        # Make sure neither the html nor the inline image is treated as an attachment:
        attachments = [value for (field, value) in files if field == "attachment"]
        self.assertEqual(len(attachments), 0)

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file:
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly:
        image = MIMEImage(image_data)
        # Mailgun requires filenames:
        image.set_param("filename", "custom-filename", "Content-Disposition")
        self.message.attach(image)

        self.message.send()
        files = self.get_api_call_files()
        attachments = [value for (field, value) in files if field == "attachment"]
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0], (image_filename, image_data, "image/png"))
        self.assertEqual(attachments[1], ("custom-filename", image_data, "image/png"))
        # Make sure the image attachments are not treated as inline:
        inlines = [value for (field, value) in files if field == "inline"]
        self.assertEqual(len(inlines), 0)

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_amp_html_alternative(self):
        # Mailgun *does* support text/x-amp-html alongside text/html
        self.message.attach_alternative("<p>HTML</p>", "text/html")
        self.message.attach_alternative("<p>And AMP HTML</p>", "text/x-amp-html")
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["html"], "<p>HTML</p>")
        self.assertEqual(data["amp-html"], "<p>And AMP HTML</p>")

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty headers"""
        self.message.send()
        data = self.get_api_call_data()
        self.assertNotIn("cc", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("h:Reply-To", data)

        # Test empty `to`--but send requires at least one recipient somewhere (like cc)
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        self.message.send()
        data = self.get_api_call_data()
        self.assertNotIn("to", data)

    def test_multiple_from_emails(self):
        """Mailgun supports multiple addresses in from_email"""
        self.message.from_email = 'first@example.com, "From, also" <second@example.com>'
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(
            data["from"], ["first@example.com", '"From, also" <second@example.com>']
        )

        # Make sure the far-more-likely scenario of a single from_email
        # with an unquoted display-name issues a reasonable error:
        self.message.from_email = "Unquoted, display-name <from@example.com>"
        with self.assertRaises(AnymailInvalidAddress):
            self.message.send()

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "Mailgun API response 400"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response:
        error_response = b"""{"message": "Helpful explanation from your ESP"}"""
        self.set_mock_response(status_code=400, raw=error_response)
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from your ESP"
        ):
            self.message.send()

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Invalid API key")
        with self.assertRaisesMessage(AnymailAPIError, "Invalid API key"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()

    def test_requests_exception(self):
        """Exception during API call should be AnymailAPIError"""
        # (The post itself raises an error--different from returning a failure response)
        from requests.exceptions import SSLError  # a low-level requests exception

        self.mock_request.side_effect = SSLError("Something bad")
        with self.assertRaisesMessage(AnymailRequestsAPIError, "Something bad") as cm:
            self.message.send()
        # also retains specific requests exception class:
        self.assertIsInstance(cm.exception, SSLError)

        # Make sure fail_silently is respected
        self.mock_request.side_effect = SSLError("Something bad")
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)


@tag("mailgun")
class MailgunBackendAnymailFeatureTests(MailgunBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_metadata(self):
        # Each metadata value is just a string; you can serialize your own JSON
        # if you'd like. (The Mailgun docs are a little confusing on this point.)
        self.message.metadata = {"user_id": "12345", "items": '["mail","gun"]'}
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["v:user_id"], "12345")
        self.assertEqual(data["v:items"], '["mail","gun"]')
        # shouldn't be needed for non-batch:
        self.assertNotIn("recipient-variables", data)

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_data()
            self.assertEqual(
                data["o:deliverytime"], "Fri, 04 Mar 2016 13:06:07 GMT"
            )  # 05:06 UTC-8 == 13:06 UTC

            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 567)
            self.message.send()
            data = self.get_api_call_data()
            self.assertEqual(
                data["o:deliverytime"], "Tue, 11 Oct 2022 06:13:14 GMT"
            )  # 12:13 UTC+6 == 06:13 UTC

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_data()
            self.assertEqual(
                data["o:deliverytime"], "Fri, 21 Oct 2022 18:00:00 GMT"
            )  # 00:00 UTC+6 == 18:00-1d UTC

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_data()
            self.assertEqual(data["o:deliverytime"], "Fri, 06 May 2022 07:08:09 GMT")

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "Thu, 13 Oct 2022 18:02:00 GMT"
            self.message.send()
            data = self.get_api_call_data()
            self.assertEqual(data["o:deliverytime"], "Thu, 13 Oct 2022 18:02:00 GMT")

    def test_tags(self):
        self.message.tags = ["receipt", "repeat-user"]
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["o:tag"], ["receipt", "repeat-user"])

    def test_tracking(self):
        # Test one way...
        self.message.track_opens = True
        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["o:tracking-opens"], "yes")
        self.assertEqual(data["o:tracking-clicks"], "no")

        # ...and the opposite way
        self.message.track_opens = False
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["o:tracking-opens"], "no")
        self.assertEqual(data["o:tracking-clicks"], "yes")

    def test_template_id(self):
        self.message.template_id = "welcome_template"
        self.message.send()
        data = self.get_api_call_data()
        self.assertEqual(data["template"], "welcome_template")

    def test_merge_data(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.body = (
            "Hi %recipient.name%. Welcome to %recipient.group% at %recipient.site%."
        )
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_global_data = {
            "group": "Users",  # default
            "site": "ExampleCo",
        }
        self.message.send()
        data = self.get_api_call_data()
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {
                    "name": "Alice",
                    "group": "Developers",
                    "site": "ExampleCo",
                },
                "bob@example.com": {
                    "name": "Bob",
                    "group": "Users",
                    "site": "ExampleCo",
                },
            },
        )
        # Make sure we didn't modify original dicts on message:
        self.assertEqual(
            self.message.merge_data,
            {
                "alice@example.com": {"name": "Alice", "group": "Developers"},
                "bob@example.com": {"name": "Bob"},
            },
        )
        self.assertEqual(
            self.message.merge_global_data, {"group": "Users", "site": "ExampleCo"}
        )

    def test_only_merge_global_data(self):
        # Make sure merge_global_data distributed to recipient-variables
        # even when merge_data not set
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_global_data = {"test": "value"}
        self.message.send()
        data = self.get_api_call_data()
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {"test": "value"},
                "bob@example.com": {"test": "value"},
            },
        )

    def test_merge_data_with_template(self):
        # Mailgun *stored* (handlebars) templates get their variable substitutions
        # from Mailgun's custom-data (not recipient-variables). To support batch sends
        # with stored templates, Anymail sets up custom-data to pull values from
        # recipient-variables. (Note this same Mailgun custom-data is also used for
        # webhook metadata tracking.)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.template_id = "welcome_template"
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_global_data = {
            "group": "Users",  # default
            "site": "ExampleCo",
        }
        self.message.send()
        data = self.get_api_call_data()
        # custom-data variables for merge_data refer to recipient-variables:
        self.assertEqual(data["v:name"], "%recipient.name%")
        self.assertEqual(data["v:group"], "%recipient.group%")
        self.assertEqual(data["v:site"], "%recipient.site%")
        # recipient-variables populates them:
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {
                    "name": "Alice",
                    "group": "Developers",
                    "site": "ExampleCo",
                },
                "bob@example.com": {
                    "name": "Bob",
                    "group": "Users",
                    "site": "ExampleCo",
                },
            },
        )

    def test_merge_metadata(self):
        # Per-recipient custom-data uses the same recipient-variables mechanism
        # as above, but prepends 'v:' to the recipient-data keys for metadata to
        # keep them separate.
        # (For on-the-fly templates -- not stored handlebars templates.)
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},
        }
        self.message.metadata = {"tier": "basic", "notification_batch": "zx912"}
        self.message.send()

        data = self.get_api_call_data()
        # custom-data variables for merge_metadata refer to recipient-variables:
        self.assertEqual(data["v:order_id"], "%recipient.v:order_id%")
        self.assertEqual(data["v:tier"], "%recipient.v:tier%")
        # metadata constant doesn't need var:
        self.assertEqual(data["v:notification_batch"], "zx912")
        # recipient-variables populates them:
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {"v:order_id": 123, "v:tier": "premium"},
                "bob@example.com": {
                    "v:order_id": 678,
                    "v:tier": "basic",  # merged from metadata default
                },
            },
        )

    def test_merge_data_with_merge_metadata(self):
        # merge_data and merge_metadata both use recipient-variables
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.body = (
            "Hi %recipient.name%. Welcome to %recipient.group% at %recipient.site%."
        )
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},  # and leave tier undefined
        }
        self.message.send()

        data = self.get_api_call_data()
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {
                    "name": "Alice",
                    "group": "Developers",
                    "v:order_id": 123,
                    "v:tier": "premium",
                },
                "bob@example.com": {
                    "name": "Bob",
                    "group": "",  # undefined merge_data --> empty string
                    "v:order_id": 678,
                    "v:tier": "",  # undefined metadata --> empty string
                },
            },
        )

    def test_merge_data_with_merge_metadata_and_template(self):
        # This case gets tricky, because when a stored template is used,
        # the per-recipient merge_metadata and merge_data both end up
        # in the same Mailgun custom-data keys.
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.template_id = "order_notification"
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123, "tier": "premium"},
            "bob@example.com": {"order_id": 678},  # and leave tier undefined
        }
        self.message.send()

        data = self.get_api_call_data()
        # custom-data covers both merge_data and merge_metadata...

        # from merge_data:
        self.assertEqual(data["v:name"], "%recipient.name%")
        self.assertEqual(data["v:group"], "%recipient.group%")

        # from merge_metadata:
        self.assertEqual(data["v:order_id"], "%recipient.v:order_id%")
        self.assertEqual(data["v:tier"], "%recipient.v:tier%")

        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {
                    "name": "Alice",
                    "group": "Developers",
                    "v:order_id": 123,
                    "v:tier": "premium",
                },
                "bob@example.com": {
                    "name": "Bob",
                    "group": "",  # undefined merge_data --> empty string
                    "v:order_id": 678,
                    "v:tier": "",  # undefined metadata --> empty string
                },
            },
        )

    def test_conflicting_merge_data_with_merge_metadata_and_template(self):
        # When a stored template is used, the same Mailgun custom-data must hold both
        # per-recipient merge_data and metadata, so there's potential for conflict.
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.template_id = "order_notification"
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},
        }
        self.message.metadata = {"group": "Order processing subsystem"}
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature,
            "conflicting merge_data and metadata keys ('group') when using template_id",
        ):
            self.message.send()

    def test_merge_headers(self):
        # Per-recipient merge_headers uses the same recipient-variables mechanism
        # as above, using variable names starting with "h:"
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            "X-Custom": "custom-default",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
                "X-No-Default": "custom-for-alice",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
                "X-Custom": "custom-for-bob",
            },
        }
        self.message.send()

        data = self.get_api_call_data()
        # non-merge header has fixed value:
        self.assertEqual(data["h:List-Unsubscribe-Post"], "List-Unsubscribe=One-Click")
        # merge headers refer to recipient-variables:
        self.assertEqual(data["h:List-Unsubscribe"], "%recipient.h:List-Unsubscribe%")
        self.assertEqual(data["h:X-Custom"], "%recipient.h:X-Custom%")
        self.assertEqual(data["h:X-No-Default"], "%recipient.h:X-No-Default%")
        # recipient-variables populates them:
        self.assertJSONEqual(
            data["recipient-variables"],
            {
                "alice@example.com": {
                    "h:List-Unsubscribe": "<https://example.com/a/>",
                    "h:X-Custom": "custom-default",  # from extra_headers
                    "h:X-No-Default": "custom-for-alice",
                },
                "bob@example.com": {
                    "h:List-Unsubscribe": "<https://example.com/b/>",
                    "h:X-Custom": "custom-for-bob",
                    "h:X-No-Default": "",  # no default in extra_headers
                },
            },
        )

    def test_force_batch(self):
        # Mailgun uses presence of recipient-variables to indicate batch send
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_data = {}
        self.message.send()
        data = self.get_api_call_data()
        self.assertJSONEqual(data["recipient-variables"], {})

    def test_sender_domain(self):
        """
        Mailgun send domain can come from from_email, envelope_sender, or esp_extra
        """
        # You could also use MAILGUN_SENDER_DOMAIN in your ANYMAIL settings,
        # as in the next test. (The mailgun_integration_tests also do that.)
        self.message.from_email = "Test From <from@from-email.example.com>"
        self.message.send()
        # API url includes the sender-domain:
        self.assert_esp_called("/from-email.example.com/messages")

        # envelope_sender (only the domain part) overrides from_email:
        self.message.from_email = "Test From <from@from-email.example.com>"
        self.message.envelope_sender = "anything@bounces.example.com"
        self.message.send()
        self.assert_esp_called("/bounces.example.com/messages")

        # esp_extra sender_domain overrides from_email:
        self.message.from_email = "Test From <from@from-email.example.com>"
        self.message.esp_extra = {"sender_domain": "esp-extra.example.com"}
        self.message.send()
        self.assert_esp_called("/esp-extra.example.com/messages")

    @override_settings(ANYMAIL_MAILGUN_SENDER_DOMAIN="mg.example.com")
    def test_sender_domain_setting(self):
        # setting overrides from_email
        self.message.send()
        self.assert_esp_called("/mg.example.com/messages")

    def test_invalid_sender_domain(self):
        # Make sure we won't construct an invalid API endpoint like
        # `https://api.mailgun.net/v3/example.com/INVALID/messages`
        # (which returns a cryptic 200-OK "Mailgun Magnificent API" response).
        self.message.from_email = "<from@example.com/invalid>"
        with self.assertRaisesMessage(
            AnymailError, "Invalid '/' in sender domain 'example.com/invalid'"
        ):
            self.message.send()

    @override_settings(ANYMAIL_MAILGUN_SENDER_DOMAIN="example.com%2Finvalid")
    def test_invalid_sender_domain_setting(self):
        # See previous test. Also, note that Mailgun unquotes % encoding *before*
        # extracting the sender domain (so %2f is just as bad as '/')
        with self.assertRaisesMessage(
            AnymailError, "Invalid '/' in sender domain 'example.com%2Finvalid'"
        ):
            self.message.send()

    @override_settings(ANYMAIL_MAILGUN_SENDER_DOMAIN="example.com # oops")
    def test_encode_sender_domain(self):
        # See previous tests. For anything other than slashes, we let Mailgun detect
        # the problem (but must properly encode the domain in the API URL)
        self.message.send()
        self.assert_esp_called("/example.com%20%23%20oops/messages")

    def test_unknown_sender_domain(self):
        self.set_mock_response(
            raw=b"""{
            "message": "Domain not found: example.com"
        }""",
            status_code=404,
        )
        with self.assertRaisesMessage(
            AnymailAPIError,
            "Unknown sender domain 'example.com'.\n"
            "Check the domain is verified with Mailgun, and that the"
            " ANYMAIL MAILGUN_API_URL setting 'https://api.mailgun.net/v3/'"
            " is the correct region.",
        ):
            self.message.send()

    @override_settings(
        # This is *not* a valid MAILGUN_API_URL setting (it should end at "...v3/"):
        ANYMAIL_MAILGUN_API_URL="https://api.mailgun.net/v3/example.com/messages"
    )
    def test_magnificent_api(self):
        # (Wouldn't a truly "magnificent API" just provide a helpful error message?)
        self.set_mock_response(raw=b"Mailgun Magnificent API", status_code=200)
        with self.assertRaisesMessage(
            AnymailAPIError,
            "Invalid Mailgun API endpoint 'https://api.mailgun.net/v3/example.com"
            "/messages/example.com/messages'.\n"
            "Check your ANYMAIL MAILGUN_SENDER_DOMAIN and MAILGUN_API_URL settings.",
        ):
            self.message.send()

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        self.assert_esp_called("/example.com/messages")
        data = self.get_api_call_data()
        mailgun_fields = {
            key: value
            for key, value in data.items()
            if key.startswith("o:") or key.startswith("v:")
        }
        self.assertEqual(mailgun_fields, {})

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        response_content = b"""{
            "id": "<12345.67890@example.com>",
            "message": "Queued. Thank you."
        }"""
        self.set_mock_response(raw=response_content)
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, "<12345.67890@example.com>")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "<12345.67890@example.com>",
        )
        self.assertEqual(msg.anymail_status.esp_response.content, response_content)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    # test_json_serialization_errors: Mailgun payload isn't JSON, so we don't test this.
    # (Anything that requests can serialize as a form field will work with Mailgun)


@tag("mailgun")
class MailgunBackendRecipientsRefusedTests(MailgunBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    # Mailgun doesn't check email bounce or complaint lists at time of send --
    # it always just queues the message. You'll need to listen for the "rejected"
    # and "failed" events to detect refused recipients.

    # The one exception is a completely invalid email, which will return a 400 response
    # and show up as an AnymailAPIError at send time.
    INVALID_TO_RESPONSE = b"""{
        "message": "'to' parameter is not a valid address. please check documentation"
    }"""

    # NOTE: As of Anymail 0.10, Anymail catches actually-invalid recipient emails before
    # attempting to pass them along to the ESP, so the tests below use technically
    # valid emails that would actually be accepted by Mailgun. (We're just making sure
    # the backend would correctly handle the 400 response if something slipped through.)

    def test_invalid_email(self):
        self.set_mock_response(status_code=400, raw=self.INVALID_TO_RESPONSE)
        msg = mail.EmailMessage(
            "Subject", "Body", "from@example.com", to=["not-really@invalid"]
        )
        with self.assertRaises(AnymailAPIError):
            msg.send()

    def test_fail_silently(self):
        self.set_mock_response(status_code=400, raw=self.INVALID_TO_RESPONSE)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["not-really@invalid"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)


@tag("mailgun")
class MailgunBackendSessionSharingTestCase(
    SessionSharingTestCases, MailgunBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("mailgun")
@override_settings(EMAIL_BACKEND="anymail.backends.mailgun.EmailBackend")
class MailgunBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        # Make sure the error mentions MAILGUN_API_KEY and ANYMAIL_MAILGUN_API_KEY
        self.assertRegex(errmsg, r"\bMAILGUN_API_KEY\b")
        self.assertRegex(errmsg, r"\bANYMAIL_MAILGUN_API_KEY\b")



import json
from datetime import datetime, timezone
from io import BytesIO
from textwrap import dedent
from unittest.mock import ANY

from django.test import override_settings, tag

from anymail.exceptions import AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.mailgun import MailgunInboundWebhookView

from .test_mailgun_webhooks import (
    TEST_WEBHOOK_SIGNING_KEY,
    mailgun_sign_legacy_payload,
    mailgun_sign_payload,
    querydict_to_postdict,
)
from .utils import (
    encode_multipart,
    make_fileobj,
    sample_email_content,
    sample_image_content,
)
from .webhook_cases import WebhookTestCase


@tag("mailgun")
@override_settings(ANYMAIL_MAILGUN_WEBHOOK_SIGNING_KEY=TEST_WEBHOOK_SIGNING_KEY)
class MailgunInboundTestCase(WebhookTestCase):
    def test_inbound_basics(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "token": "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0",
                "timestamp": "1461261330",
                "recipient": "test@inbound.example.com",
                "sender": "envelope-from@example.org",
                "message-headers": json.dumps(
                    [
                        [
                            "X-Mailgun-Spam-Rules",
                            "DKIM_SIGNED, DKIM_VALID, DKIM_VALID_AU, ...",
                        ],
                        ["X-Mailgun-Dkim-Check-Result", "Pass"],
                        ["X-Mailgun-Spf", "Pass"],
                        ["X-Mailgun-Sscore", "1.7"],
                        ["X-Mailgun-Sflag", "No"],
                        ["X-Mailgun-Incoming", "Yes"],
                        ["X-Envelope-From", "<envelope-from@example.org>"],
                        ["Received", "from mail.example.org by mxa.mailgun.org ..."],
                        [
                            "Received",
                            "by mail.example.org for <test@inbound.example.com> ...",
                        ],
                        [
                            "Dkim-Signature",
                            "v=1; a=rsa-sha256; c=relaxed/relaxed; d=example.org; ...",
                        ],
                        ["Mime-Version", "1.0"],
                        [
                            "Received",
                            "by 10.10.1.71 with HTTP;"
                            " Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
                        ],
                        ["From", '"Displayed From" <from+test@example.org>'],
                        ["Date", "Wed, 11 Oct 2017 18:31:04 -0700"],
                        ["Message-Id", "<CAEPk3R+4Zr@mail.example.org>"],
                        ["Subject", "Test subject"],
                        [
                            "To",
                            '"Test Inbound" <test@inbound.example.com>,'
                            " other@example.com",
                        ],
                        ["Cc", "cc@example.com"],
                        [
                            "Content-Type",
                            'multipart/mixed; boundary="089e0825ccf874a0bb055b4f7e23"',
                        ],
                    ]
                ),
                "body-plain": "Test body plain",
                "body-html": "<div>Test body html</div>",
                "stripped-html": "stripped html body",
                "stripped-text": "stripped plaintext body",
            }
        )
        response = self.client.post("/anymail/mailgun/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailgunInboundWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 21, 17, 55, 30, tzinfo=timezone.utc)
        )
        self.assertEqual(
            event.event_id, "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0"
        )
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(querydict_to_postdict(event.esp_event.POST), raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "Test body plain")
        self.assertEqual(message.html, "<div>Test body html</div>")

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertEqual(message.stripped_text, "stripped plaintext body")
        self.assertEqual(message.stripped_html, "stripped html body")
        self.assertIs(message.spam_detected, False)
        self.assertEqual(message.spam_score, 1.7)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by mxa.mailgun.org ...",
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

    def test_attachments(self):
        att1 = BytesIO("test attachment".encode("utf-8"))
        att1.name = "test.txt"
        image_content = sample_image_content()
        att2 = BytesIO(image_content)
        att2.name = "image.png"
        email_content = sample_email_content()
        att3 = BytesIO(email_content)
        att3.name = "\\share\\mail\\forwarded.msg"
        att3.content_type = 'message/rfc822; charset="us-ascii"'
        raw_event = mailgun_sign_legacy_payload(
            {
                "message-headers": "[]",
                "attachment-count": "3",
                "content-id-map": """{"<abc123>": "attachment-2"}""",
                "attachment-1": att1,
                "attachment-2": att2,  # inline
                "attachment-3": att3,
            }
        )

        response = self.client.post("/anymail/mailgun/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailgunInboundWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "test attachment")
        # Django strips paths:
        self.assertEqual(attachments[1].get_filename(), "forwarded.msg")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)

    def test_filtered_attachment_filenames(self):
        # Make sure the inbound webhook can deal with missing fields caused by
        # Django's multipart/form-data filename filtering. (The attachments are lost,
        # but shouldn't cause errors in the inbound webhook.)
        filenames = [
            "",
            "path\\",
            "path/" ".",
            "path\\.",
            "path/.",
            "..",
            "path\\..",
            "path/..",
        ]
        num_attachments = len(filenames)
        payload = {
            "attachment-%d"
            % (i + 1): make_fileobj(
                "content", filename=filenames[i], content_type="text/pdf"
            )
            for i in range(num_attachments)
        }
        payload.update(
            {
                "message-headers": "[]",
                "attachment-count": str(num_attachments),
            }
        )

        # Must do our own multipart/form-data encoding for empty filenames:
        response = self.client.post(
            "/anymail/mailgun/inbound/",
            data=encode_multipart("BoUnDaRy", mailgun_sign_legacy_payload(payload)),
            content_type="multipart/form-data; boundary=BoUnDaRy",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailgunInboundWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )

        # Different Django releases strip different filename patterns.
        # Just verify that at least some attachments got dropped (so the test is valid)
        # without causing an error in the inbound webhook:
        attachments = kwargs["event"].message.attachments
        self.assertLess(len(attachments), num_attachments)

    def test_unusual_content_id_map(self):
        # Under unknown conditions, Mailgun appears to generate a content-id-map with
        # multiple empty keys (and possibly other duplicate keys). We still want to
        # correctly identify inline attachments from it.
        raw_event = mailgun_sign_legacy_payload(
            {
                "message-headers": "[]",
                "attachment-count": "4",
                "content-id-map": '{"": "attachment-1", "": "attachment-2",'
                ' "<abc>": "attachment-3", "<abc>": "attachment-4"}',
                "attachment-1": make_fileobj("att1"),
                "attachment-2": make_fileobj("att2"),
                "attachment-3": make_fileobj("att3"),
                "attachment-4": make_fileobj("att4"),
            }
        )

        response = self.client.post("/anymail/mailgun/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailgunInboundWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        message = event.message
        self.assertEqual(len(message.attachments), 0)  # all inlines
        inlines = [part for part in message.walk() if part.is_inline()]
        self.assertEqual(len(inlines), 4)
        self.assertEqual(inlines[0]["Content-ID"], "")
        self.assertEqual(inlines[1]["Content-ID"], "")
        self.assertEqual(inlines[2]["Content-ID"], "<abc>")
        self.assertEqual(inlines[3]["Content-ID"], "<abc>")

    def test_inbound_mime(self):
        # Mailgun provides the full, raw MIME message if the webhook url ends in 'mime'
        raw_event = mailgun_sign_legacy_payload(
            {
                "token": "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0",
                "timestamp": "1461261330",
                "recipient": "test@inbound.example.com",
                "sender": "envelope-from@example.org",
                "body-mime": dedent(
                    """\
                From: A tester <test@example.org>
                Date: Thu, 12 Oct 2017 18:03:30 -0700
                Message-ID: <CAEPk3RKEx@mail.example.org>
                Subject: Raw MIME test
                To: test@inbound.example.com
                MIME-Version: 1.0
                Content-Type: multipart/alternative; boundary="94eb2c05e174adb140055b6339c5"

                --94eb2c05e174adb140055b6339c5
                Content-Type: text/plain; charset="UTF-8"
                Content-Transfer-Encoding: quoted-printable

                It's a body=E2=80=A6

                --94eb2c05e174adb140055b6339c5
                Content-Type: text/html; charset="UTF-8"
                Content-Transfer-Encoding: quoted-printable

                <div dir=3D"ltr">It's a body=E2=80=A6</div>

                --94eb2c05e174adb140055b6339c5--
                """  # NOQA: E501
                ),
            }
        )

        response = self.client.post("/anymail/mailgun/inbound_mime/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailgunInboundWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        message = event.message
        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertEqual(message.subject, "Raw MIME test")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\n""",
        )

    def test_misconfigured_tracking(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "clicked",
                    "timestamp": 1534109600.089676,
                    "recipient": "recipient@example.com",
                    "url": "https://example.com/test",
                }
            }
        )
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set Mailgun's *clicked tracking* webhook"
            " to Anymail's Mailgun *inbound* webhook URL.",
        ):
            self.client.post(
                "/anymail/mailgun/inbound/",
                data=json.dumps(raw_event),
                content_type="application/json",
            )

    def test_misconfigured_store_action(self):
        # store() notification includes "attachments" json;
        # forward() includes "attachment-count"
        raw_event = mailgun_sign_legacy_payload(
            {
                "token": "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0",
                "timestamp": "1461261330",
                "recipient": "test@inbound.example.com",
                "sender": "envelope-from@example.org",
                "body-plain": "Test body plain",
                "body-html": "<div>Test body html</div>",
                "attachments": json.dumps(
                    [
                        {
                            "url": "https://storage.mailgun.net/v3/domains/example.com"
                            "/messages/MESSAGE_KEY/attachments/0",
                            "content-type": "application/pdf",
                            "name": "attachment.pdf",
                            "size": 20202,
                        }
                    ]
                ),
            }
        )
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have configured Mailgun's receiving route using the store()"
            " action. Anymail's inbound webhook requires the forward() action.",
        ):
            self.client.post("/anymail/mailgun/inbound/", data=raw_event)

    def test_misconfigured_tracking_legacy(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "domain": "example.com",
                "message-headers": "[]",
                "recipient": "recipient@example.com",
                "event": "delivered",
            }
        )
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set Mailgun's *delivered tracking* webhook"
            " to Anymail's Mailgun *inbound* webhook URL.",
        ):
            self.client.post("/anymail/mailgun/inbound/", data=raw_event)
